import jax
from flax import nnx


class ConvBlock(nnx.Module):
    def __init__(self, in_features: int, out_features: int, num_conv: int, rngs: nnx.Rngs):
        layers = []
        for _ in range(num_conv):
            conv = nnx.Conv(in_features, out_features, kernel_size=(3, 3), padding=1, rngs=rngs)
            layers.append(conv)
            layers.append(nnx.relu)
            in_features = out_features
        self.conv = nnx.Sequential(*layers)

    def __call__(self, x: jax.Array) -> jax.Array:
        return self.conv(x)


class VGG16Features(nnx.Module):
    def __init__(self, rngs: nnx.Rngs) -> None:
        self.conv1 = ConvBlock(3, 64, num_conv=2, rngs=rngs)
        self.conv2 = ConvBlock(64, 128, num_conv=2, rngs=rngs)
        self.conv3 = ConvBlock(128, 256, num_conv=3, rngs=rngs)
        self.conv4 = ConvBlock(256, 512, num_conv=3, rngs=rngs)
        self.conv5 = ConvBlock(512, 512, num_conv=3, rngs=rngs)

    def __call__(self, x: jax.Array) -> list[jax.Array]:
        act1 = self.conv1(x)
        act2 = self.conv2(nnx.max_pool(act1, window_shape=(2, 2), strides=(2, 2)))
        act3 = self.conv3(nnx.max_pool(act2, window_shape=(2, 2), strides=(2, 2)))
        act4 = self.conv4(nnx.max_pool(act3, window_shape=(2, 2), strides=(2, 2)))
        act5 = self.conv5(nnx.max_pool(act4, window_shape=(2, 2), strides=(2, 2)))

        return [act1, act2, act3, act4, act5]



import jax
from flax import nnx


class AlexNetFeatures(nnx.Module):
    def __init__(self, rngs: nnx.Rngs) -> None:
        self.conv1 = nnx.Conv(3, 64, kernel_size=(11, 11), strides=4, padding=2, rngs=rngs)
        self.conv2 = nnx.Conv(64, 192, kernel_size=(5, 5), padding=2, rngs=rngs)
        self.conv3 = nnx.Conv(192, 384, kernel_size=(3, 3), padding=1, rngs=rngs)
        self.conv4 = nnx.Conv(384, 256, kernel_size=(3, 3), padding=1, rngs=rngs)
        self.conv5 = nnx.Conv(256, 256, kernel_size=(3, 3), padding=1, rngs=rngs)

    def __call__(self, x: jax.Array) -> list[jax.Array]:
        act1 = nnx.relu(self.conv1(x))
        act2 = nnx.relu(self.conv2(nnx.max_pool(act1, window_shape=(3, 3), strides=(2, 2))))
        act3 = nnx.relu(self.conv3(nnx.max_pool(act2, window_shape=(3, 3), strides=(2, 2))))
        act4 = nnx.relu(self.conv4(act3))
        act5 = nnx.relu(self.conv5(act4))

        return [act1, act2, act3, act4, act5]



from typing import Callable, TypeVar

import jax
from flax import nnx
from flax.traverse_util import unflatten_dict
from huggingface_hub import hf_hub_download
from safetensors.flax import load_file

from jaxlpips.modules import AlexNetFeatures, VGG16Features

ModelType = TypeVar("ModelType", bound=nnx.Module)


def load_model(model_name: str) -> tuple[AlexNetFeatures | VGG16Features, list[jax.Array]]:
    match model_name:
        case "alexnet":
            model_class = AlexNetFeatures
        case "vgg16":
            model_class = VGG16Features
        case _:
            raise ValueError("Only 'alexnet' and 'vgg16' pretrained networks are supported.")

    model_weights_file = hf_hub_download(
        repo_id="tux-type/jaxlpips", filename=model_name + "_features.safetensors"
    )
    model_weights = load_file(model_weights_file)
    linear_weights_file = hf_hub_download(
        repo_id="tux-type/jaxlpips", filename=model_name + "_lpips.safetensors"
    )
    linear_weights = list(load_file(linear_weights_file).values())

    model = to_nnx(model_weights, lambda: model_class(rngs=nnx.Rngs(0)))
    return model, linear_weights


def to_nnx(flat_weights: dict[str, jax.Array], lazy_model: Callable[[], ModelType]) -> ModelType:
    graph_def = nnx.graphdef(nnx.eval_shape(lazy_model))
    flat_dict = {tuple(k.split(".")): v for k, v in flat_weights.items()}
    state_dict = unflatten_dict(flat_dict)
    model = nnx.merge(graph_def, state_dict)
    return model



import lpips
import numpy as np
import pytest
import torch

from jaxlpips import LPIPS as JaxLPIPS


def to_torch_tensor(x: np.ndarray):
    out = x.transpose((0, 3, 1, 2))
    out = torch.tensor(out, requires_grad=False)
    return out


@pytest.fixture(scope="session")
def dummy_data():
    rng = np.random.default_rng(2781)

    ref = rng.normal(size=(4, 64, 64, 3)).astype(np.float32)
    tgt = rng.normal(size=(4, 64, 64, 3)).astype(np.float32)
    return ref, tgt


@pytest.fixture(scope="session")
def img_data():
    rng = np.random.default_rng(9127)
    ref = np.load("tests/data/test_imgs.npy")
    tgt = ref + rng.normal(0, 0.5, size=ref.shape).astype(np.float32)
    return ref, tgt


@pytest.mark.parametrize("pretrained_network", ["alex", "vgg"])
def test_lpips_dummy(dummy_data, pretrained_network):
    ref, tgt = dummy_data
    jax_loss_fn = JaxLPIPS(pretrained_network=pretrained_network)
    loss_fn = lpips.LPIPS(net=pretrained_network)

    jax_output = jax_loss_fn(ref, tgt)
    torch_output = loss_fn(to_torch_tensor(ref), to_torch_tensor(tgt)).detach().cpu().numpy()

    assert np.allclose(jax_output, torch_output, rtol=0, atol=1e-5)


@pytest.mark.parametrize("pretrained_network", ["alex", "vgg"])
def test_lpips_img(img_data, pretrained_network):
    ref, tgt = img_data
    jax_loss_fn = JaxLPIPS(pretrained_network=pretrained_network)
    loss_fn = lpips.LPIPS(net=pretrained_network)

    jax_output = jax_loss_fn(ref, tgt)
    torch_output = loss_fn(to_torch_tensor(ref), to_torch_tensor(tgt)).detach().cpu().numpy()

    assert np.allclose(jax_output, torch_output, rtol=0, atol=1e-5)



import jax
import jax.numpy as jnp
import numpy as np

from jaxlpips.utils import load_model


class LPIPS:
    def __init__(self, pretrained_network: str = "alexnet"):
        if pretrained_network == "alex":
            pretrained_network = "alexnet"
        if pretrained_network == "vgg":
            pretrained_network = "vgg16"

        valid_networks = ["alexnet", "vgg16"]
        if pretrained_network not in valid_networks:
            raise ValueError("Only 'alexnet' and 'vgg16' pretrained networks are supported.")

        self.pretrained_network = pretrained_network

        self.model, self.linear_weights = load_model(self.pretrained_network)

    def __call__(self, ref: jax.Array, tgt: jax.Array):
        ref_feats, tgt_feats = self.model(scale_shift(ref)), self.model(scale_shift(tgt))

        layer_dists = [
            jnp.mean(
                jnp.sum(((normalise(ref_ft) - normalise(tgt_ft)) ** 2) * w, axis=3, keepdims=True),
                axis=(1, 2),
                keepdims=True,
            )
            for ref_ft, tgt_ft, w in zip(ref_feats, tgt_feats, self.linear_weights, strict=True)
        ]
        dist_score = np.sum(layer_dists, axis=0)
        return dist_score


def scale_shift(x: jax.Array):
    shift = jnp.array([-0.030, -0.088, -0.188])[None, None, None, :]
    scale = jnp.array([0.458, 0.448, 0.450])[None, None, None, :]
    return (x - shift) / scale


def normalise(x: jax.Array, eps=1e-10):
    norm_factor = jnp.sqrt(jnp.sum(x**2, axis=3, keepdims=True))
    return x / (norm_factor + eps)
